@@ -32,8 +32,6 @@
  *   <img src="toast-overlay.png" alt="toast-overlay">
  * </picture>
  *
- * Much like [class@Gtk.Overlay], `AdwToastOverlay` is a container with a single
- * main child, on top of which it can display a [class@Toast], overlaid.
  * Toasts can be shown with [method@ToastOverlay.add_toast].
  *
  * See [class@Toast] for details.
@@ -44,7 +42,9 @@
  * toastoverlay
  * ├── [child]
  * ├── toast
- * ┊   ├── label.heading
+ * ┊   ├── widget
+ * ┊   │   ├── [label.heading]
+ *     │   ╰── [custom title]
  *     ├── [button]
  *     ╰── button.circular.flat
  * ```
@@ -52,9 +52,12 @@
  * `AdwToastOverlay`'s CSS node is called `toastoverlay`. It contains the child,
  * as well as zero or more `toast` subnodes.
  *
- * Each of the `toast` nodes contains a `label` subnode with the `.heading`
- * style class, optionally a `button` subnode, and another `button` subnode with
- * `.circular` and `.flat` style classes.
+ * Each of the `toast` nodes contains a `widget` subnode, optionally a `button`
+ * subnode, and another `button` subnode with `.circular` and `.flat` style
+ * classes.
+ *
+ * The `widget` subnode contains a `label` subnode with the `.heading` style
+ * class, or a custom widget provided by the application.
  *
  * ## Accessibility
  *
@@ -217,6 +220,7 @@ dismissed_cb (ToastInfo *info)
   } else {
     g_queue_remove (self->queue, info);
 
+    adw_toast_set_overlay (ADW_TOAST (info->toast), NULL);
     if (!info->hide_animation)
       free_toast_info (info);
   }
@@ -256,6 +260,51 @@ show_toast (AdwToastOverlay *self,
   adw_animation_play (info->show_animation);
 }
 
+static int
+bump_sort_func (ToastInfo *compare,
+                AdwToast  *toast,
+                gpointer   user_data)
+{
+  if (adw_toast_get_priority (compare->toast) == ADW_TOAST_PRIORITY_HIGH)
+    return -1;
+
+  return 1;
+}
+
+static int
+find_toast_func (ToastInfo *info,
+                 AdwToast  *toast)
+{
+  if (info && info->toast == toast)
+    return 0;
+
+  return 1;
+}
+
+static void
+bump_toast (AdwToastOverlay *self,
+            AdwToast        *toast)
+{
+  GList *link;
+  ToastInfo *info;
+
+  /* Remove it from the queue, then reinsert in the right location */
+  link = g_queue_find_custom (self->queue, toast,
+                              (GCompareFunc) find_toast_func);
+
+  g_assert (link);
+
+  info = link->data;
+  g_queue_remove (self->queue, info);
+
+  if (adw_toast_get_priority (toast) == ADW_TOAST_PRIORITY_HIGH)
+    g_queue_push_head (self->queue, info);
+  else
+    g_queue_insert_sorted (self->queue, info,
+                           (GCompareDataFunc) bump_sort_func,
+                           NULL);
+}
+
 static gboolean
 dismiss_cb (AdwToastOverlay *self,
             GVariant        *args)
@@ -317,7 +366,11 @@ allocate_toast (AdwToastOverlay *self,
 
   gtk_widget_get_preferred_size (info->widget, NULL, &size);
 
-  size.width = MIN (MAX (size.width, NATURAL_WIDTH), width);
+  if (adw_toast_widget_get_button_visible (ADW_TOAST_WIDGET (info->widget)))
+    size.width = MIN (MAX (size.width, NATURAL_WIDTH), width);
+  else
+    size.width = MIN (size.width, width);
+
   size.height = MIN (size.height, height);
 
   x = (width - size.width) / 2;
@@ -449,11 +502,9 @@ adw_toast_overlay_class_init (AdwToastOverlayClass *klass)
    * Since: 1.0
    */
   props[PROP_CHILD] =
-      g_param_spec_object ("child",
-                           "Child",
-                           "The child widget",
-                           GTK_TYPE_WIDGET,
-                           G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY);
+    g_param_spec_object ("child", NULL, NULL,
+                         GTK_TYPE_WIDGET,
+                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
@@ -569,6 +620,11 @@ adw_toast_overlay_set_child (AdwToastOverlay *self,
  * either @toast or the original toast will be placed in a queue, depending on
  * the priority of @toast. See [property@Toast:priority].
  *
+ * If called on a toast that's already displayed, its timeout will be reset.
+ *
+ * If called on a toast currently in the queue, the toast will be bumped
+ * forward to be shown as soon as possible.
+ *
  * Since: 1.0
  */
 void
@@ -576,18 +632,35 @@ adw_toast_overlay_add_toast (AdwToastOverlay *self,
                              AdwToast        *toast)
 {
   ToastInfo *info;
+  AdwToastOverlay *overlay;
 
   g_return_if_fail (ADW_IS_TOAST_OVERLAY (self));
   g_return_if_fail (ADW_IS_TOAST (toast));
 
-  if (adw_toast_get_added (toast)) {
-    g_critical ("Adding toast '%s', but it has already been added to an "
-                "AdwToastOverlay", adw_toast_get_title (toast));
+  overlay = adw_toast_get_overlay (toast);
 
+  /* If the toast has been added already and is being shown, reset its
+   * timeout. Otherwise, bump it forward in the queue. */
+  if (overlay == self) {
+    if (self->current_toast && self->current_toast->toast == toast)
+      adw_toast_widget_reset_timeout (ADW_TOAST_WIDGET (self->current_toast->widget));
+    else
+      bump_toast (self, toast);
+
+    /* Unref, as we don't actually take the ownership */
+    g_object_unref (toast);
     return;
   }
 
-  adw_toast_set_added (toast, TRUE);
+  if (overlay) {
+    g_critical ("Adding toast '%s', but it has already been added to a "
+                "different AdwToastOverlay", adw_toast_get_title (toast));
+
+    g_object_unref (toast);
+    return;
+  }
+
+  adw_toast_set_overlay (toast, self);
 
   info = g_new0 (ToastInfo, 1);
   info->overlay = self;
